import logging
import sys
import os
from textwrap import dedent

import requests
from tambo import Transport

import chacractl
from chacractl.util import retry

logger = logging.getLogger(__name__)


class Project(object):
    _help = dedent("""
    Handle projects on a remote chacra instance.

    Creating a new project::

        chacractl project create project

    Options:

    create        Creates a new project
    """)
    help_menu = "create projects"
    options = ['create']

    def __init__(self, argv):
        self.argv = argv

    @property
    def base_url(self):
        return os.path.join(
            chacractl.config['url'], 'binaries'
        )

    def sanitize_url(self, url_part):
        # get rid of the leading slash to prevent issues when joining
        url = url_part.lstrip('/')

        # and add a trailing slash so that the request is done at the correct
        # canonical url
        if not url.endswith('/'):
            url = "%s/" % url
        return url

    @retry()
    def post(self, url):
        exists = requests.head(url, verify=chacractl.config['ssl_verify'])

        if exists.status_code == 200:
            logger.warning('resource exists, will not upload')
            logger.warning('SKIP %s', url)
            return
        elif exists.status_code == 404:
            logger.info('POSTing to project: %s', url)
            response = requests.post(
                url,
                auth=chacractl.config['credentials'],
                verify=chacractl.config['ssl_verify'])
        if response.status_code > 201:
            logger.warning("%s -> %s", response.status_code, response.text)
            response.raise_for_status()

    @retry()
    def delete(self, url):
        # XXX This exists here but it is not yet implemented, e.g. nothing
        # calls this method
        exists = requests.head(url, verify=chacractl.config['ssl_verify'])
        if exists.status_code == 404:
            logger.warning('project already deleted')
            logger.warning('SKIP %s', url)
            return
        logger.info('DELETE project: %s', url)
        response = requests.delete(
            url,
            auth=chacractl.config['credentials'],
            verify=chacractl.config['ssl_verify'])
        if response.status_code > 201:
            logger.warning("%s -> %s", response.status_code, response.text)

    def main(self):
        self.parser = Transport(self.argv, options=self.options)
        self.parser.catch_help = self._help
        self.parser.parse_args()

        # handle posting projects:
        if self.parser.has('create'):
            url_part = self.sanitize_url(self.parser.get('create'))
            if not sys.stdin.isatty():
                # read from stdin
                logger.info('reading input from stdin')
                for line in sys.stdin.readlines():
                    url = os.path.join(self.base_url, url_part)
                    self.post(url)
            else:
                url = os.path.join(self.base_url, url_part)
                self.post(url)
        # XXX this exists here but it not yet enabled from the CLI
        elif self.parser.has('delete'):
            url_part = self.sanitize_url(self.parser.get('delete'))
            url = os.path.join(self.base_url, url_part)
            self.delete(url)



import imp
import os
from textwrap import dedent
from requests.exceptions import BaseHTTPError, RequestException

import logging

logger = logging.getLogger(__name__)


def default_config_path():
    return os.path.expanduser(u'~/.chacractl')


def get_config_path():
    """
    Return the .chacractl configuration location
    """
    config = default_config_path()
    if os.path.exists(config):
        return config


def load_config():
    config = get_config_path()
    if not config:
        return
    return imp.load_source('chacractl', config)


def ensure_default_config():
    """
    Create a default config if the config file does not currently exist
    """
    template = dedent("""
    # This file was automatically generated by the chacractl CLI
    # make sure to update it with the correct user and key to talk to the API
    url = "http://example"
    user = "chacra user"
    key = "secret"
    ssl_verify = True
    """)
    config = default_config_path()
    if not os.path.exists(config):
        with open(config, 'w') as f:
            f.write(template)

import functools
import time

def retry(exceptions=(RequestException, BaseHTTPError), delay=5, times=3):
    """
    A decorator for retrying a function call with a specified delay in case of a set of exceptions, mainly
    intended to avoid the *horrendous* requests lacking API to deal with this.

    :param exceptions:  A tuple of all exceptions that need to be caught for retry
                                        e.g. retry(exception_list = (Timeout, Readtimeout))
    :param delay: Amount of delay (seconds) needed between successive retries.
    :param times: no of times the function should be retried
    """
    def outer_wrapper(function):
        @functools.wraps(function)
        def inner_wrapper(*args, **kwargs):
            final_excep = None
            for counter in xrange(times):
                if counter > 0:
                    time.sleep(delay)
                final_excep = None
                try:
                    value = function(*args, **kwargs)
                    return value
                except (exceptions) as e:
                    final_excep = e
                    if hasattr(e, 'response'):
                        if hasattr(e.response, 'status_code'):
                            status = e.response.status_code
                            if status < 500:
                                raise
                    logger.warning('while trying a request, got an exception: %s', str(e))
                    pass

            if final_excep is not None:
                raise final_excep
        return inner_wrapper

    return outer_wrapper



import os
from textwrap import dedent
import pytest
import chacractl
from chacractl.main import ChacraCtl


class TestApiCredentials(object):

    def chacractl_templates():
        template_base = dedent("""
        url = "https://example"
        user = "chacra user"
        key = "secret"
        """)
        return [
            (template_base, True),
            (template_base + 'ssl_verify = True', True),
            (template_base + 'ssl_verify = False', False),
            (template_base + 'ssl_verify = "path/to/ca"', 'path/to/ca')
        ]

    @pytest.mark.parametrize('template,expected', chacractl_templates())
    def test_ssl_verify(self, tmpdir, monkeypatch, template, expected):
        config_path = tmpdir.join('.chacractl')
        config_path.write(template)
        monkeypatch.setattr(os.path, 'expanduser', lambda _: str(config_path))
        c = ChacraCtl()
        c.api_credentials()
        assert chacractl.config['ssl_verify'] == expected
